import time
from typing import Union

import requests

from .constants import COPI_API, PRICE_API
from .exceptions import LoginError


def login(self, password) -> None:
    """Attempt to login to the API."""
    self._public_ip = public_ip()

    body = {"customer": self.username, "password": password}

    try:
        r = requests.post(
            f"{PRICE_API}/authenticate",
            data=body,
            headers=headers(self),
        )
        r.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise LoginError(f"Failed to login: {e}")

    json = r.json()

    self._token = json["token"]
    self._zip_code = json["address"]["zip_code"]


def init_vars(self) -> None:
    """Set all the variables required for all other functions."""
    r = requests.get(f"{COPI_API}/profile", headers=headers(self))
    r.raise_for_status()

    json = r.json()

    self.motivation = json["primaryMotivation"]
    self.accommodation_type = json["accommodationType"]

    # grid_area is only mentioned in forward-prices...
    r = requests.get(
        f"{PRICE_API}/forward-prices/{self._zip_code}",
        headers=headers(self),
    )
    r.raise_for_status()

    json = r.json()
    self._grid_area = json["0"]["grid_area"]


def headers(self) -> dict:
    """Set appropiate headers for API calls."""
    return {
        "X-Customer-Ip": self._public_ip,
        "Authorization": self._token,
    }


def public_ip() -> str:
    """Return the public ip."""
    return requests.get("https://api.ipify.org").text


def add_ints_avg(raw_list: list) -> set:
    """Return the average of `raw_list`, trying to estimate if necessary.

    :param raw_list: a list of numbers and "N/A"s
    :type raw_list: list
    :rtype: set
    """
    estimate = False
    num = 0
    str_count = [type(n) for n in raw_list].count(str)

    if str_count > 0:
        estimate = True
        non_str_count = len(raw_list) - str_count

        # Add the average for each "N/A" in the list. Incase we
        # only have strings, just add 0. We have no idea
        if non_str_count > 0:
            num = sum(n for n in raw_list if type(n) != str)
            num += (num / non_str_count) * str_count

            num = num / len(raw_list)
        else:
            num = 0.0

    else:
        num = sum(raw_list) / len(raw_list)

    return (round(num, 2), estimate)


def get_timestamp(t: Union[str, int]) -> str:
    """
    Return `t` or the timestamp at hour `t` if only an hour was provided.

    :param t: a timestamp or an hour
    :type t: int | str
    :rtype: str
    """
    if type(t) == int or (type(t) == str and t.isnumeric()):
        t = int(t)

        if t > 24:
            day = int(time.strftime("%d")) + 1
            hour = t - 24

            return time.strftime(f"%Y-%m-{str(day).zfill(2)}T{str(hour).zfill(2)}:00:00")
        else:
            return time.strftime(f"%Y-%m-%dT{str(t).zfill(2)}:00:00")

    return t



"""Main module."""


class PySeasNVE:
    from .core import (
        best_interval,
        cheapest_interval,
        climate_at,
        current_co2_intensity,
        current_green_energy,
        current_price,
        forecast_climate,
        forecast_price,
        greenest_interval,
        price_at,
    )

    def __init__(self, username, password):
        from .helpers import init_vars, login

        self.username = username

        # Private vars
        self._token = ""
        self._grid_area = ""
        self._zip_code = ""
        self._public_ip = ""
        self._cached_forecast_price = {}
        self._cached_forecast_climate = {}

        # Properties
        self._accommodation_type = ""
        self._motivation = ""

        # Login instantly, skipping saving the plain text password in self
        login(self, password)

        # Preload variables
        init_vars(self)

    @property
    def motivation(self):
        return self._motivation

    @motivation.setter
    def motivation(self, m):
        if m in ["economy", "environment", "both"]:
            self._motivation = m
        else:
            raise ValueError("motivation can only be one of `economy`, `environment`, or `both`")

    @property
    def accommodation_type(self):
        return self._accommodation_type

    @accommodation_type.setter
    def accommodation_type(self, t):
        if t in ["apartment", "house"]:
            self._accommodation_type = t
        else:
            raise ValueError("accommodation_type can only be `apartment` or `house`")


class DummyPySeasNVE:
    """A dummy class for unit testing"""

    from .core import (
        best_interval,
        cheapest_interval,
        climate_at,
        current_co2_intensity,
        current_green_energy,
        current_price,
        forecast_climate,
        forecast_price,
        greenest_interval,
        price_at,
    )

    def __init__(self, username, password):
        self.username = username

        self._token = "dummy_token"
        self._grid_area = "DK-2"
        self._zip_code = "2000"
        self._public_ip = "127.0.0.1"
        self._accommodation_type = "apartment"
        self._motivation = "both"
        self._cached_forecast_climate = {}
        self._cached_forecast_price = {}



import time
from math import ceil, floor
from typing import Union

from .forecast import climate, price
from .helpers import add_ints_avg, get_timestamp


def forecast_price(self) -> dict:
    """Return the price"""
    return price(self)


def forecast_climate(self) -> dict:
    """Return the climate"""
    return climate(self)


def current_price(self, type: str = "total") -> float:
    """Return the current price (by default the total price) in DKK/kWh.

    :param self: self
    :param type: one of `total` (default), `raw_price`, or `tax`
    :type type: str
    :rtype: float
    """
    if type not in ["total", "raw_price", "tax"]:
        raise ValueError("Type is not one of `total`, `raw_price`, or `tax`")

    prices = price(self)
    return prices[int(time.strftime("%H"))][f"kwh_{type}"]


def current_green_energy(self) -> float:
    """Return the current green energy in percent.

    :param self: self
    :rtype: float
    """
    climates = climate(self)
    return climates[int(time.strftime("%H"))]["green_energy_percent"]


def current_co2_intensity(self) -> int:
    """Return the current co2 intensity in gCO2eq/kWh.

    :param self: self
    :rtype: int
    """
    climates = climate(self)
    return climates[int(time.strftime("%H"))]["co2_intensity"]


def price_at(self, timestamp: Union[str, int]) -> dict:
    """Return the price at `timestamp`.

    :param self: self
    :param timestamp: a timestamp in the format "2022-03-19T08:00:00"
        or an hour "8" or "36"
    :type timestamp: str | int
    :rtype: dict
    """
    prices = price(self)

    timestamp = get_timestamp(timestamp)

    for v in prices.values():
        if v["start_time"] == timestamp:
            return v

    return {}


def climate_at(self, timestamp: Union[str, int]) -> dict:
    """Return the climate at `timestamp`.

    :param self: self
    :param timestamp: a timestamp in the format "2022-03-19T08:00:00"
        or an hour "8" or "36"
    :type timestamp: str | int
    :rtype: dict
    """
    climates = climate(self)

    timestamp = get_timestamp(timestamp)

    for v in climates.values():
        if v["start_time"] == timestamp:
            return v

    return {}


# for *_interval we should consider apartments
# that can't wash during the night
def best_interval(self, interval: int = 1, items: int = 3) -> list:
    """Return the greenest/cheapest/mixed interval(s) depending on primary motivation.

    :param self: self
    :param interval: how many continuous hours to calculate
    :type interval: int
    :param items: how many items to return
    :type items: int
    :rtype: list
    """
    if self.motivation == "economy":
        return cheapest_interval(self, interval, items)
    elif self.motivation == "both":
        cheapest = cheapest_interval(self, interval, items)
        greenest = greenest_interval(self, interval, items)

        uniq = list()
        for i in cheapest + greenest:
            if i not in uniq:
                uniq.append(i)

        # We can still have a green bias ;-)
        n_green = ceil(items / 2)
        n_cheap = floor(items / 2)

        return sorted(
            uniq[0:n_green] + uniq[n_green : n_green + n_cheap],
            key=lambda x: (
                x["interval_avg_green_energy_percent_estimate"],  # prefer non-estimates
                x["interval_avg_kwh_price_estimate"],
                -x["interval_avg_green_energy_percent"],
                x["interval_avg_kwh_price"],
            ),
        )
    else:
        return greenest_interval(self, interval, items)


def cheapest_interval(self, interval: int = 1, items: int = 3) -> list:
    """Return the cheapest intervals.

    :param self: self
    :param interval: how many continuous hours to calculate
    :type interval: int
    :param items: how many items to return
    :type items: int
    :rtype: list
    """
    prices = price(self)

    cheap_keys = list()

    for key in prices.keys():
        interval_price = 0
        interval_energy = list()

        try:
            for i in range(interval):
                interval_price += prices[key + i]["kwh_total"]
                interval_energy.append(
                    climate_at(self, prices[key + i]["start_time"]).get("green_energy_percent", "N/A")
                )
        except KeyError:
            break

        energy, estimate = add_ints_avg(interval_energy)

        cheap_keys.append(
            {
                "start_time": prices[key]["start_time"],
                "interval_hours": interval,
                "interval_avg_kwh_price": round(interval_price / interval, 2),
                "interval_avg_kwh_price_estimate": False,
                "interval_avg_green_energy_percent": energy,
                "interval_avg_green_energy_percent_estimate": estimate,
            }
        )

    return sorted(
        cheap_keys,
        key=lambda x: (
            x["interval_avg_kwh_price"],
            x["interval_avg_green_energy_percent_estimate"],  # prefer non-estimates
            -x["interval_avg_green_energy_percent"],
        ),
    )[0:items]


def greenest_interval(self, interval: int = 1, items: int = 3) -> list:
    """Return the greenest intervals.

    :param self: self
    :param interval: how many continuous hours to calculate
    :type interval: int
    :param items: how many items to return
    :type items: int
    :rtype: list
    """
    climates = climate(self)

    green_keys = list()

    for key in climates.keys():
        interval_energy = 0
        interval_price = list()

        try:
            for i in range(interval):
                interval_energy += climates[key + i]["green_energy_percent"]
                interval_price.append(price_at(self, climates[key + i]["start_time"]).get("kwh_total", "N/A"))
        except KeyError:
            break

        price, estimate = add_ints_avg(interval_price)

        green_keys.append(
            {
                "start_time": climates[key]["start_time"],
                "interval_hours": interval,
                "interval_avg_kwh_price": price,
                "interval_avg_kwh_price_estimate": estimate,
                "interval_avg_green_energy_percent": round(interval_energy / interval, 2),
                "interval_avg_green_energy_percent_estimate": False,
            }
        )

    return sorted(
        green_keys,
        key=lambda x: (
            -x["interval_avg_green_energy_percent"],
            x["interval_avg_kwh_price_estimate"],  # prefer non-estimates
            x["interval_avg_kwh_price"],
        ),
    )[0:items]



import datetime
import time

import requests
from pytz import timezone

from .constants import CLIMATE_API, KEY_TIMESTAMP_FMT, PRICE_API
from .helpers import headers


def is_cached(cache_obj: dict) -> bool:
    """Check if `cache_obj` is cached.

    :param cache_obj: a intance variable where data could be cached
    :type cache_obj: dict
    :rtype: bool
    """
    if len(cache_obj) > 0:
        hour = datetime.datetime.now().hour
        if hour == 0 and cache_obj["cached_hour"] != 0:
            return False
        else:
            return True
    else:
        return False


def price(self) -> dict:
    """Get the price forecast."""
    if is_cached(self._cached_forecast_price):
        return self._cached_forecast_price["data"]

    r = requests.get(
        f"{PRICE_API}/forward-prices/{self._zip_code}",
        headers=headers(self),
    )
    json = r.json()

    # dict.$HOUR is electric prices without taxes
    # dict.distribution_prices are taxes, but only for today (?)
    prices = {}
    tax_prices = json["distribution_prices"][0]["prices"]

    for i in range(48):
        o = json[str(i)]
        t = datetime.datetime.strptime(o["valid_from"], "%Y-%m-%d %H:%M:%S")

        mday = int(time.strftime("%d"))
        hour = int(time.strftime("%H"))
        if t.day < mday or (t.day == mday and t.hour < hour):
            # A forecast doesn't account for past time
            continue

        if i >= 24:
            # The tax price only has hours for today (?); it looks like
            # they are reusing the prices again for tomorrow
            kwh_tax = round(tax_prices[i - 24]["price"]) / 100
        else:
            kwh_tax = round(tax_prices[i]["price"]) / 100

        kwh_raw_price = round(o["price"]) / 100
        kwh_total = round(kwh_raw_price + kwh_tax, 2)

        # The API only updates prices every other day it seems.
        # Correct in case we start at 2nd day of prices
        if i >= 24 and i - 24 not in prices.keys():
            key = i - 24
        else:
            key = i

        prices[key] = {
            "start_time": t.strftime(KEY_TIMESTAMP_FMT),
            "kwh_raw_price": kwh_raw_price,
            "kwh_tax": kwh_tax,
            "kwh_total": kwh_total,
        }

    # "cache" the result, as it only changes every hour 00:00
    self._cached_forecast_price = {
        "cached_hour": datetime.datetime.now().hour,
        "data": prices,
    }

    return prices


def climate(self) -> dict:
    """Get the climate forecast."""
    if is_cached(self._cached_forecast_climate):
        return self._cached_forecast_climate["data"]

    r = requests.get(
        f"{CLIMATE_API}/energydata/DK-{self._grid_area}",
        headers=headers(self),
    )
    json = r.json()

    sorted_climate = sorted(json, key=lambda x: x["start"])
    climates = {}

    for i in sorted_climate:
        # For climates, they output dates in GMT+0 (one hour behind)
        t = datetime.datetime.strptime(i["start"], "%Y-%m-%dT%H:%M:%S.%fZ")

        # So we should correct it to apply Danish time
        t = t.astimezone(timezone("Europe/Copenhagen"))
        offset = t.utcoffset().seconds
        t = t + datetime.timedelta(seconds=offset)

        mday = int(time.strftime("%d"))
        hour = int(time.strftime("%H")) - offset / 3600  # also correct time.now()
        if t.day < mday or (t.day == mday and t.hour < hour):
            # A forecast doesn't account for past time
            continue

        c_biomass = i["powerConsumptionBreakdown"]["biomass"] or 0
        c_coal = i["powerConsumptionBreakdown"]["coal"] or 0
        c_gas = i["powerConsumptionBreakdown"]["gas"] or 0
        c_geothermal = i["powerConsumptionBreakdown"]["geothermal"] or 0
        c_hydro = i["powerConsumptionBreakdown"]["hydro"] or 0
        c_nuclear = i["powerConsumptionBreakdown"]["nuclear"] or 0
        c_oil = i["powerConsumptionBreakdown"]["oil"] or 0
        c_solar = i["powerConsumptionBreakdown"]["solar"] or 0
        c_wind = i["powerConsumptionBreakdown"]["wind"] or 0
        c_unknown = i["powerConsumptionBreakdown"]["unknown"] or 0

        # This seems to be what they qualify as green/black energy
        c_green = c_biomass + c_geothermal + c_hydro + c_nuclear + c_solar + c_wind
        c_total = c_green + c_coal + c_gas + c_oil + c_unknown

        c_green_energy_percent = round(c_green / c_total * 100, 2)

        if t.day > int(time.strftime("%d")):
            key = t.hour + 24
        else:
            key = t.hour

        climates[key] = {
            "start_time": t.strftime(KEY_TIMESTAMP_FMT),
            "green_energy_percent": c_green_energy_percent,
            "co2_intensity": i["co2_intensity"],
            "consumption_breakdown_percent": {
                "biomass": round(c_biomass / c_total * 100, 2),
                "coal": round(c_coal / c_total * 100, 2),
                "gas": round(c_gas / c_total * 100, 2),
                "geothermal": round(c_geothermal / c_total * 100, 2),
                "hydro": round(c_hydro / c_total * 100, 2),
                "nuclear": round(c_nuclear / c_total * 100, 2),
                "oil": round(c_oil / c_total * 100, 2),
                "solar": round(c_solar / c_total * 100, 2),
                "wind": round(c_wind / c_total * 100, 2),
                "unknown": round(c_unknown / c_total * 100, 2),
            },
        }

    # "cache" the result, as it only changes every hour 00:00
    self._cached_forecast_climate = {
        "cached_hour": datetime.datetime.now().hour,
        "data": climates,
    }

    return climates



#!/usr/bin/env python

"""Tests for `pyseasnve` package."""


import unittest

from pyseasnve import DummyPySeasNVE


class TestPyseasnve(unittest.TestCase):
    """Tests for `pyseasnve` package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.seas = DummyPySeasNVE("test", "test")
        self.seas._cached_forecast_price = {
            "cached_hour": 13,
            "data": {
                13: {"start_time": "2022-03-20T13:00:00", "kwh_raw_price": 0.01, "kwh_tax": 1.56, "kwh_total": 1.57},
                14: {"start_time": "2022-03-20T14:00:00", "kwh_raw_price": 0.18, "kwh_tax": 1.59, "kwh_total": 1.77},
                15: {"start_time": "2022-03-20T15:00:00", "kwh_raw_price": 0.18, "kwh_tax": 2.01, "kwh_total": 2.19},
            },
        }
        self.seas._cached_forecast_climate = {
            "cached_hour": 13,
            "data": {
                13: {
                    "start_time": "2022-03-20T13:00:00",
                    "green_energy_percent": 85.68,
                    "co2_intensity": 204,
                    "consumption_breakdown_percent": {
                        "biomass": 18.64,
                        "coal": 16.9,
                        "gas": 5.41,
                        "geothermal": 0.0,
                        "hydro": 13.09,
                        "nuclear": 10.02,
                        "oil": 0.0,
                        "solar": 3.94,
                        "wind": 29.99,
                        "unknown": 2.0,
                    },
                },
                14: {
                    "start_time": "2022-03-20T14:00:00",
                    "green_energy_percent": 75.68,
                    "co2_intensity": 182,
                    "consumption_breakdown_percent": {
                        "biomass": 18.64,
                        "coal": 16.9,
                        "gas": 5.41,
                        "geothermal": 0.0,
                        "hydro": 13.09,
                        "nuclear": 10.02,
                        "oil": 0.0,
                        "solar": 3.94,
                        "wind": 29.99,
                        "unknown": 2.0,
                    },
                },
                15: {
                    "start_time": "2022-03-20T15:00:00",
                    "green_energy_percent": 80.68,
                    "co2_intensity": 194,
                    "consumption_breakdown_percent": {
                        "biomass": 18.64,
                        "coal": 16.9,
                        "gas": 5.41,
                        "geothermal": 0.0,
                        "hydro": 13.09,
                        "nuclear": 10.02,
                        "oil": 0.0,
                        "solar": 3.94,
                        "wind": 29.99,
                        "unknown": 2.0,
                    },
                },
            },
        }

    def tearDown(self):
        """Tear down test fixtures, if any."""
        del self.seas

    def test_greenest_1_interval(self):
        """Test the greenest_interval with interval=1"""
        expected_result = [
            {
                "start_time": "2022-03-20T13:00:00",
                "interval_hours": 1,
                "interval_avg_kwh_price": 1.57,
                "interval_avg_kwh_price_estimate": False,
                "interval_avg_green_energy_percent": 85.68,
                "interval_avg_green_energy_percent_estimate": False,
            },
            {
                "start_time": "2022-03-20T15:00:00",
                "interval_hours": 1,
                "interval_avg_kwh_price": 2.19,
                "interval_avg_kwh_price_estimate": False,
                "interval_avg_green_energy_percent": 80.68,
                "interval_avg_green_energy_percent_estimate": False,
            },
        ]
        actual_result = self.seas.greenest_interval(1, 2)

        self.assertListEqual(actual_result, expected_result)

    def test_greenest_2_interval(self):
        """Test the greenest_interval function with interval=2"""
        expected_result = [
            {
                "start_time": "2022-03-20T13:00:00",
                "interval_hours": 2,
                "interval_avg_kwh_price": 1.67,
                "interval_avg_kwh_price_estimate": False,
                "interval_avg_green_energy_percent": 80.68,
                "interval_avg_green_energy_percent_estimate": False,
            },
            {
                "start_time": "2022-03-20T14:00:00",
                "interval_hours": 2,
                "interval_avg_kwh_price": 1.98,
                "interval_avg_kwh_price_estimate": False,
                "interval_avg_green_energy_percent": 78.18,
                "interval_avg_green_energy_percent_estimate": False,
            },
        ]
        actual_result = self.seas.greenest_interval(2, 2)

        self.assertListEqual(actual_result, expected_result)

    def test_cheapest_1_interval(self):
        """Test the cheapest_interval with interval=1"""
        expected_result = [
            {
                "start_time": "2022-03-20T13:00:00",
                "interval_hours": 1,
                "interval_avg_kwh_price": 1.57,
                "interval_avg_kwh_price_estimate": False,
                "interval_avg_green_energy_percent": 85.68,
                "interval_avg_green_energy_percent_estimate": False,
            },
            {
                "start_time": "2022-03-20T14:00:00",
                "interval_hours": 1,
                "interval_avg_kwh_price": 1.77,
                "interval_avg_kwh_price_estimate": False,
                "interval_avg_green_energy_percent": 75.68,
                "interval_avg_green_energy_percent_estimate": False,
            },
        ]
        actual_result = self.seas.cheapest_interval(1, 2)

        self.assertListEqual(actual_result, expected_result)

    def test_cheapest_2_interval(self):
        """Test the cheapest_interval with interval=2"""
        expected_result = [
            {
                "start_time": "2022-03-20T13:00:00",
                "interval_hours": 2,
                "interval_avg_kwh_price": 1.67,
                "interval_avg_kwh_price_estimate": False,
                "interval_avg_green_energy_percent": 80.68,
                "interval_avg_green_energy_percent_estimate": False,
            },
            {
                "start_time": "2022-03-20T14:00:00",
                "interval_hours": 2,
                "interval_avg_kwh_price": 1.98,
                "interval_avg_kwh_price_estimate": False,
                "interval_avg_green_energy_percent": 78.18,
                "interval_avg_green_energy_percent_estimate": False,
            },
        ]
        actual_result = self.seas.cheapest_interval(2, 2)

        self.assertListEqual(actual_result, expected_result)

    def test_climate_at_timestamp(self):
        """Test the climate_at function, using a timestamp"""
        expected_result = {
            "start_time": "2022-03-20T15:00:00",
            "green_energy_percent": 80.68,
            "co2_intensity": 194,
            "consumption_breakdown_percent": {
                "biomass": 18.64,
                "coal": 16.9,
                "gas": 5.41,
                "geothermal": 0.0,
                "hydro": 13.09,
                "nuclear": 10.02,
                "oil": 0.0,
                "solar": 3.94,
                "wind": 29.99,
                "unknown": 2.0,
            },
        }
        actual_result = self.seas.climate_at("2022-03-20T15:00:00")

        self.assertDictEqual(actual_result, expected_result)

    def test_price_at_timestamp(self):
        """Test the price_at function, using a timestamp"""
        expected_result = {
            "start_time": "2022-03-20T14:00:00",
            "kwh_raw_price": 0.18,
            "kwh_tax": 1.59,
            "kwh_total": 1.77,
        }
        actual_result = self.seas.price_at("2022-03-20T14:00:00")

        self.assertDictEqual(actual_result, expected_result)
